/***************************************************************************/
/*                                                                         */
/*  DiffCon.c                                                              */
/*             Temperature Difference Control with LCD                     */
/*             PIC16F877  4Mhz                                             */
/*                                                                         */
/*      Mode : HD44780 type LCD displays                                   */
/*                                                                         */
/* Interface :     SEL = Port D 4                                          */
/*                  RS = Port D 5                                          */
/*              Data_7 = Port D 3                                          */
/*              Data_6 = Port D 2                                          */
/*              Data_5 = Port D 1                                          */
/*              Data_4 = Port D 0                                          */
/*              Solar  = Port A 0  // Temperature Solar boiler             */
/*              Heat   = Port A 1  // Temperature Heater boiler            */
/*               VR+   = Port A 3  // +5V reference voltage                */
/*    Taster1 (ROT)    = Port B 1  // Taster1 Input (Mode,Set)             */
/*    Taster2 (SCHWARZ)= Port B 2  // Taster2 Input (+)                    */
/*    Taster3 (SCHWARZ)= Port B 3  // Taster3 Input (-)                    */
/*          PCF583 SDA = Port C 4  // RealTimeClock IC per I2C Bus         */
/*          PCF583 SCL = Port C 3  //       "                              */
/*              valve  = Port C 5  //                                      */
/*              pump   = Port C 6  //                                      */
/*                                                                         */
/* (c) Matthias Stehr 18.08.2001                                           */
/*                                                                         */
/*                                                                         */
/***************************************************************************/
#include <16F877.h>
 #use i2c(MASTER, SDA=PIN_C4, SCL=PIN_C3)

//#define DEBUG


#ifdef DEBUG
#use delay(clock=20000000)         // Crystal on Dev. board
#else
#use delay(clock=4000000)          // Crystal on target system
#endif // DEBUG


#include <lcd_d.c>    // LCD an Port D
 

#define PortBConfig  0b11111111   /* Configure port B  0 = output       */
#define PortCConfig  0b10000111   /* Configure port C  0 = output       */
#define EE_DTEMPON   0x00         // EE-Prom address
#define EE_DTEMPOFF  0x01
#define EE_RUNTIMEL  0x02         // Runtime of the pump low  byte
#define EE_RUNTIMEH  0x03         // Runtime of the pump high byte
#define EE_MAINTON   0x04         // Maintenance Feature on/off
#define EE_HHMAINT   0x05         // Maintenance Starttime hh
#define EE_MMMAINT   0x06         // Maintenance Stoptime  mm

#bit ADFM = 0x9F.7     // A/D Result format

void messen(void);
void control(void);
void keys(void);
void ReadTime(void);
void WriteTime(void);
char BcdToChar(char c);
char CharToBcd(char b);
void time(void);
void CalculateRunTime(void);
void ReadRunTime(void);
void SaveRunTime(void);
void InputRunTime(void);
void InputMaintenance(void);
void InputMaintTime(void);

long SolarTemp, HeaterTemp, messwert, x,
     dTempOn, dTempOff, RunTime;
signed long DeltaTemp;
char i, ss, mm, hh, mmTemp, mmRunTime, MaintenanceOn,
     mmMaint, hhMaint;
boolean IsPumpOn;


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void main()
{
   set_tris_b( PortBConfig );
   set_tris_c( PortCConfig );
   output_low(PIN_C5);
   output_low(PIN_C6);
   #ifdef DEBUG
   setup_adc_ports(RA0_RA1_RA3_ANALOG);
   #else
   setup_adc_ports(RA0_RA1_ANALOG_RA3_REF);
   #endif //DEBUG
   setup_adc(ADC_CLOCK_INTERNAL);
   ADFM = TRUE;

   ss         = 40;
   mm         = 59;
   hh         = 19;
   IsPumpOn   = FALSE;
   SolarTemp  = 0;
   HeaterTemp = 0;
   DeltaTemp  = 0;
   mmTemp     = 0;
   mmRunTime  = 0;
   dTempOn    = read_eeprom(EE_DTEMPON);
   dTempOff   = read_eeprom(EE_DTEMPOFF);
   ReadRunTime();
   hhMaint = read_eeprom(EE_HHMAINT);
   mmMaint = read_eeprom(EE_MMMAINT);
   MaintenanceOn = read_eeprom(EE_MAINTON);

   #ifdef DEBUG
   WriteTime();
   #endif //DEBUG
   lcd_init();
   printf(lcd_putc,"\fDiffCon V1.2\n(c) M.Stehr 2005");
   delay_ms(2000);
   printf(lcd_putc,"\f");
   do
      {
      messen();
      ReadTime();
      lcd_gotoxy(1,1);
      printf(lcd_putc,"Th:%2lu", HeaterTemp); lcd_putc(0xDF); lcd_putc('C');
      printf(lcd_putc," Ts:%2lu",SolarTemp);lcd_putc(0xDF); lcd_putc('C');
      printf(lcd_putc,"\n%02u:%02u dT:%2ld", hh, mm, DeltaTemp);
      if(IsPumpOn)
         printf(lcd_putc," AN  ");
      else
         printf(lcd_putc," AUS ");
      control();
      if(MaintenanceOn)
         time();                // Check if Maintenance should be done
      keys();                   // Check user input
      }
   while(1);
}




/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void messen(void)
{
   set_adc_channel(0);
   delay_us(20);
   x = 0;
   for(i= 0; i<16; i++)
      {
      messwert = read_adc();
      x+= messwert;
      }
   messwert = x/16;
   SolarTemp = messwert/10;

   set_adc_channel(1);
   delay_us(20);
   x = 0;
   for(i= 0; i<16; i++)
      {
      messwert = read_adc();
      x+= messwert;
      }
   messwert = x/16;
   HeaterTemp = messwert/10;

   DeltaTemp = SolarTemp - HeaterTemp;
}





/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void control(void)
{
   if(DeltaTemp >= dTempOn)
      {
      if(IsPumpOn == FALSE)
         {
         mmTemp = mm; // Store the current minute
         }
      IsPumpOn = TRUE;
      output_high(PIN_C5); // pump on
      delay_ms(2000);          // wait 2 sec
      output_high(PIN_C6); // valve open
      CalculateRunTime();
      }
   if(DeltaTemp <= dTempOff)
      {
      IsPumpOn = FALSE;
      output_low(PIN_C6);  // valve close
      delay_ms(2000);          // wait 2 sec
      output_low(PIN_C5);  // pump off
      }
}





/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void keys(void)
{
   if( !input(PIN_B1) )
      {
      delay_ms(10);
      if( !input(PIN_B1) )
         {
         printf(lcd_putc,"\f<EINSTELLUNGEN>");
         while( !input(PIN_B1) )
            ;
         /*****************************************************************/
         InputRunTime();
         /*****************************************************************/
         InputMaintenance();
         /*****************************************************************/
         if(MaintenanceOn)
            InputMaintTime();
         /*****************************************************************/
         printf(lcd_putc,"\f<deltaTemp AN>");
         while(input(PIN_B1))
            {
            lcd_gotoxy(1,2);
            printf(lcd_putc," %2lu", dTempOn);
            if( !input(PIN_B2) )
               {
               delay_ms(10);
               if( !input(PIN_B2) )
                  {
                  while( !input(PIN_B2) )
                     ;
                  dTempOn= dTempOn + 1;
                  if(dTempOn > 20)
                     dTempOn = 20;
                  }
               }
            if( !input(PIN_B3) )
               {
               delay_ms(10);
               if( !input(PIN_B3) )
                  {
                  while( !input(PIN_B3) )
                     ;
                  dTempOn= dTempOn - 1;
                  if(dTempOn < 2)
                     dTempOn = 2;
                  }
               }
            }
         while( !input(PIN_B1) )
            ;
         /*****************************************************************/
         printf(lcd_putc,"\f<deltaTemp AUS>");
         while(input(PIN_B1))
            {
            lcd_gotoxy(1,2);
            printf(lcd_putc," %2lu", dTempOff);
            if( !input(PIN_B2) )
               {
               delay_ms(10);
               if( !input(PIN_B2) )
                  {
                  while( !input(PIN_B2) )
                     ;
                  dTempOff= dTempOff + 1;
                  if(dTempOff > 10)
                     dTempOff = 10;
                  }
               }
            if( !input(PIN_B3) )
               {
               delay_ms(10);
               if( !input(PIN_B3) )
                  {
                  while( !input(PIN_B3) )
                     ;
                  dTempOff= dTempOff - 1;
                  if(dTempOff < 1)
                     dTempOff = 1;
                  }
               }
            }
         while( !input(PIN_B1) )
            ;
         /*****************************************************************/
         printf(lcd_putc,"\f<TIME: hh>");
         while(input(PIN_B1))
            {
            lcd_gotoxy(1,2);
            printf(lcd_putc,"\n%02u:%02u  ", hh, mm);
            if( !input(PIN_B2) )
               {
               delay_ms(10);
               if( !input(PIN_B2) )
                  {
                  while( !input(PIN_B2) )
                     ;
                  hh= hh + 1;
                  if(hh > 23)
                     hh = 0;
                  }
               }
            if( !input(PIN_B3) )
               {
               delay_ms(10);
               if( !input(PIN_B3) )
                  {
                  while( !input(PIN_B3) )
                     ;
                  hh= hh - 1;
                  if(hh > 23)
                     hh = 0;
                  }
               }
            }
         while( !input(PIN_B1) )
            ;
         /*****************************************************************/
         printf(lcd_putc,"\f<TIME:   mm>");
         while(input(PIN_B1))
            {
            lcd_gotoxy(1,2);
            printf(lcd_putc,"\n%02u:%02u  ", hh, mm);
            if( !input(PIN_B2) )
               {
               delay_ms(10);
               if( !input(PIN_B2) )
                  {
                  while( !input(PIN_B2) )
                     ;
                  mm= mm + 1;
                  if(mm > 59)
                     mm = 0;
                  }
               }
            if( !input(PIN_B3) )
               {
               delay_ms(10);
               if( !input(PIN_B3) )
                  {
                  while( !input(PIN_B3) )
                     ;
                  mm= mm - 1;
                  if(mm > 59)
                     mm = 0;
                  }
               }
            }
         while( !input(PIN_B1) )
            ;
         ss = 0;
         WriteTime();
         write_eeprom(EE_DTEMPON,  dTempOn);
         write_eeprom(EE_DTEMPOFF, dTempOff);
         SaveRunTime();
         write_eeprom(EE_MAINTON, MaintenanceOn);
         /*****************************************************************/
         }
      }
}




/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void InputRunTime(void)
{
  printf(lcd_putc,"\fRunTime:%5luh\n", RunTime);
  printf(lcd_putc,"<Set><  ><Reset>", dTempOn);
  while(input(PIN_B1))
     {
     lcd_gotoxy(1,1);
     printf(lcd_putc,"RunTime:%5lu", RunTime);
     if( !input(PIN_B2) )
        {
        delay_ms(10);
        if( !input(PIN_B2) )
           {
           while( !input(PIN_B2) )
              ;
           ;
           }
        }
     if( !input(PIN_B3) )
        {
        delay_ms(10);
        if( !input(PIN_B3) )
           {
           while( !input(PIN_B3) )
              ;
           RunTime=0;
           }
        }
     }
  while( !input(PIN_B1) )
     ;
}




/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void InputMaintenance(void)
{
  printf(lcd_putc,"\fWartungsfunktion");
  while(input(PIN_B1))
     {
     lcd_gotoxy(1,2);
     if(MaintenanceOn)
       printf(lcd_putc,"AN ");
     else
       printf(lcd_putc,"AUS");
     if( !input(PIN_B2) )
        {
        delay_ms(10);
        if( !input(PIN_B2) )
           {
           while( !input(PIN_B2) )
              ;
           MaintenanceOn = FALSE;
           }
        }
     if( !input(PIN_B3) )
        {
        delay_ms(10);
        if( !input(PIN_B3) )
           {
           while( !input(PIN_B3) )
              ;
           MaintenanceOn = TRUE;
           }
        }
     }
  while( !input(PIN_B1) )
     ;
}



/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void InputMaintTime(void)
{
  printf(lcd_putc,"\f<WART.START:hh>");
  while(input(PIN_B1))
     {
     lcd_gotoxy(1,2);
     printf(lcd_putc,"\n%02u:%02u  ", hhMaint, mmMaint);
     if( !input(PIN_B2) )
        {
        delay_ms(10);
        if( !input(PIN_B2) )
           {
           while( !input(PIN_B2) )
              ;
           hhMaint= hhMaint + 1;
           if(hhMaint > 23)
              hhMaint = 0;
           }
        }
     if( !input(PIN_B3) )
        {
        delay_ms(10);
        if( !input(PIN_B3) )
           {
           while( !input(PIN_B3) )
              ;
           hhMaint= hhMaint - 1;
           if(hhMaint > 23)
              hhMaint = 0;
           }
        }
     }
  while( !input(PIN_B1) )
     ;
  /*****************************************************************/
  printf(lcd_putc,"\f<WART.START:  mm>");
  while(input(PIN_B1))
     {
     lcd_gotoxy(1,2);
     printf(lcd_putc,"\n%02u:%02u  ", hhMaint, mmMaint);
     if( !input(PIN_B2) )
        {
        delay_ms(10);
        if( !input(PIN_B2) )
           {
           while( !input(PIN_B2) )
              ;
           mmMaint= mmMaint + 1;
           if(mmMaint > 59)
              mmMaint = 0;
           }
        }
     if( !input(PIN_B3) )
        {
        delay_ms(10);
        if( !input(PIN_B3) )
           {
           while( !input(PIN_B3) )
              ;
           mmMaint= mmMaint - 1;
           if(mmMaint > 59)
              mmMaint = 0;
           }
        }
     }
  while( !input(PIN_B1) )
     ;

  write_eeprom(EE_HHMAINT,  hhMaint);
  write_eeprom(EE_MMMAINT,  mmMaint);
}










/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void ReadTime(void)
{
  i2c_Start();               // Set a START condition I2C Bus
  i2c_Write( 0b10100010 );   // Address and Write Flag
  i2c_Write( 0b00000010 );   // read from Byte 2
  i2c_Start();               // Set a START condition I2C Bus
  i2c_Write( 0b10100011 );   // Address and Read Flag
  ss = BcdToChar(i2c_Read());  // sec
  mm = BcdToChar(i2c_Read());  // min
  hh = BcdToChar(i2c_Read(0)); // hours and NO_ACK;
  i2c_stop();
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void WriteTime(void)
{
  i2c_Start();                  // Set a START condition I2C Bus
  i2c_Write( 0b10100010 );      // Address and Write Flag
  i2c_Write( 0b00000010 );      // write from Byte 2
  i2c_Write(CharToBcd(ss));     // sec
  i2c_Write(CharToBcd(mm));     // min
  i2c_Write(CharToBcd(hh));     // hh
  i2c_stop();
}


/**************************************************************************/
/*                                                                        */
/**************************************************************************/
char BcdToChar(char c)
{
   return ((c/16)*10 + (c%16));
}




/**************************************************************************/
/*                                                                        */
/**************************************************************************/
char CharToBcd(char b)
{
   return((b/10)*16 + (b%10));
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void time(void)
{
   if((hh == hhMaint) && (mm == mmMaint))
      {
      // in the winter time the pump should run minimum 1min per day
      // it protects the pump and valve hanging.
      printf(lcd_putc,"\fWartungslauf \nPumpe an");
      output_high(PIN_C5); // pump on
      delay_ms(5000);      // wait 5 sec
      printf(lcd_putc,"\fWartungslauf \nVentil auf");
      output_high(PIN_C6); // valve open
      delay_ms(5000);
      x= 40;
      do
         {
         delay_ms(1000);     // wait 1 sec
         lcd_gotoxy(1,2);
         printf(lcd_putc,"noch %2lu sec", x);
         x = x - 1;
         }
      while(x != 0);
      printf(lcd_putc,"\fWartungslauf \nVentil zu");
      output_low(PIN_C6);  // valve close
      delay_ms(5000);      // wait 2 sec
      output_low(PIN_C5);  // pump off
      printf(lcd_putc,"\fWartungslauf \nPumpe aus");
      delay_ms(5000);
      }
}






/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void CalculateRunTime(void)
{
   // Note: This Runtime calculation is not 100% exact,
   // but for the solar control it's good enough.
   if(mmTemp != mm)
      {                 // one minute is gone
      ++mmRunTime;
      mmTemp = mm;
      if(mmRunTime == 60)
         {
         // one hour is gone and should be added to the
         // EE-Prom RunTime
         RunTime = RunTime + 1;
         SaveRunTime();        // Store the runtime in EE-Prom;
         mmRunTime = 0;
         }
      }
}





/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void ReadRunTime(void)
{
   *(&RunTime + 0) = read_eeprom(EE_RUNTIMEL);
   *(&RunTime + 1) = read_eeprom(EE_RUNTIMEH);
}





/**************************************************************************/
/*                                                                        */
/**************************************************************************/
void SaveRunTime(void)
{
   write_eeprom(EE_RUNTIMEL,  *(&Runtime + 0));   // Low Byte
   write_eeprom(EE_RUNTIMEH,  *(&Runtime + 1));   // High Byte
}




